package ru.langrafhomyak.db.jdbc_resources_manager;

import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/* package */ class _SynchronizedEmbedKeyHashMap<T extends _SynchronizedEmbedKeyHashMap.Node<T>> {
	public static abstract class Node<T extends _SynchronizedEmbedKeyHashMap.Node<T>> {
		public final int keyHashCode;
		@Nullable
		/*private*/ T _nextNode;

		public Node(int keyHashCode) {
			this.keyHashCode = keyHashCode;
		}

		protected abstract boolean _checkKeyReference(@NotNull Object ref);
	}


	@NotNull
	private final Int2ObjectOpenHashMap<T> _impl;
	@NotNull
	private final Object _sync;

	public _SynchronizedEmbedKeyHashMap() {
		this._impl = new Int2ObjectOpenHashMap<>();
		this._sync = new Object();
	}

	@Nullable
	public T get(int keyHashCode, @NotNull Object ref) {
		synchronized (this._sync) {
			T p = this._impl.get(keyHashCode);
			while (p != null) {
				if (p._checkKeyReference(ref))
					return p;
				p = p._nextNode;
			}
		}
		return null;
	}

	public void add(@NotNull T node) {
		synchronized (this._sync) {
			node._nextNode = this._impl.put(node.keyHashCode, node);
		}
	}

	public void remove(@NotNull T node) {
		synchronized (this._sync) {
			REMOVED:
			{
				T p = this._impl.get(node.keyHashCode);
				if (p == node) {
					this._impl.put(node.keyHashCode, p._nextNode);
					break REMOVED;
				}

				while (p != null) {
					T n = p._nextNode;
					if (n == node) {
						p._nextNode = n._nextNode;
						break REMOVED;
					}
					p = n;
				}
				throw new IllegalArgumentException("Node not found");
			}
			this._impl.trim();
		}
	}
}
