package ru.langrafhomyak.db.jdbc_resources_manager

import java.lang.ref.PhantomReference
import java.lang.ref.ReferenceQueue

internal object _PhantomCacheCleanup {
	private val _thread: Thread

	@Suppress("MAY_BE_CONST")
	private val _virtualThreadProperty = "ru.langrafhomyak.db.jdbc_resources_manager._PhantomCacheCleanup.USE_VIRTUAL_THREAD"
	private val _referenceQueue = ReferenceQueue<Any?>()

	@Suppress("MAY_BE_CONST")
	private val _threadName = "ru.langrafhomyak.db.jdbc_resources_manager._PhantomCacheCleanup"


	init {
		if (this._isVirtualThreadsSupported() && System.getProperty(this._virtualThreadProperty)?.lowercase() != "false")
			this._thread = this._startVirtualThread()
		else
			this._thread = this._startSystemThreadBeforeJ21()
	}

	private fun _isVirtualThreadsSupported(): Boolean {
		try {
			Thread::ofVirtual
			return true
		} catch (_: NoSuchMethodError) {
			return false
		}
	}

	private fun _startSystemThreadBeforeJ21(): Thread {
		val t = Thread(
			null,
			CleanupRoutine, this._threadName,
			1024 * 16, false
		)
		t.isDaemon = true
		t.priority = Thread.MIN_PRIORITY + 1
		t.start()
		return t
	}

	private fun _startVirtualThread(): Thread {
		val b = Thread.ofVirtual()
		b.name(this._threadName)
		return b.start(CleanupRoutine)
	}


	private object CleanupRoutine : Runnable {
		override fun run() {
			while (true) {
				// (Thread.currentThread() as? InnocuousThread)?.eraseThreadLocals()

				try {
					(this@_PhantomCacheCleanup._referenceQueue.remove(60 * 60) as CleanablePhantomReference<*>).cleanup()
				} /* catch (_: LinkageError) {
					return
				} */ catch (_: InterruptedException) {
					return
				} catch (e: Throwable) {
					e.printStackTrace()
				}
			}

		}
	}

	abstract class CleanablePhantomReference<T>(ref: T) : PhantomReference<T>(ref, this@_PhantomCacheCleanup._referenceQueue) {
		abstract fun cleanup()
	}
}